/*
 * Copyright 2016 The OpenSSL Project Authors. All Rights Reserved.
 *
 * Licensed under the Apache License 2.0 (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://www.openssl.org/source/license.html
 */

/*
 * This file is only used by HP C on VMS, and is included automatically
 * after each header file from this directory
 */

/* restore state.  Must correspond to the save in __decc_include_prologue.h */
#pragma names restore
