#! /usr/bin/env perl
# Copyright 2015-2026 The OpenSSL Project Authors. All Rights Reserved.
#
# Licensed under the Apache License 2.0 (the "License").  You may not use
# this file except in compliance with the License.  You can obtain a copy
# in the file LICENSE in the source distribution or at
# https://www.openssl.org/source/license.html


use OpenSSL::Test;
use OpenSSL::Test::Utils;
use OpenSSL::Test::Simple;

setup("test_abort");

plan skip_all => "This test should not be run under valgrind"
    if ( defined $ENV{OSSL_USE_VALGRIND} );

plan tests => 1;

is(run(test(["aborttest"])), 0, "Testing that abort is caught correctly");
