#! /usr/bin/env perl
# Copyright 2015-2026 The OpenSSL Project Authors. All Rights Reserved.
#
# Licensed under the Apache License 2.0 (the "License").  You may not use
# this file except in compliance with the License.  You can obtain a copy
# in the file LICENSE in the source distribution or at
# https://www.openssl.org/source/license.html


use OpenSSL::Test;
use OpenSSL::Test::Utils;

setup("test_memleak");

plan skip_all => "This test should not be run under valgrind"
    if (defined ($ENV{OSSL_USE_VALGRIND}));

plan skip_all => "MacOS currently doesn't support leak sanitizer"
    if $^O eq 'darwin';

plan tests => 2;
ok(!run(test(["memleaktest"])), "running leak test");
ok(run(test(["memleaktest", "freeit"])), "running no leak test");
